from ..Auth import Auth
from ..Config import APIConfig
from ..models.Hit import Hit
from ..models.HitType import HitType
from ..models.Season import Season
from ..utils.parser import parse_participants, parse_hits


class Series(Hit):

    def __init__(self, title, series_id, hit_type: HitType):
        super().__init__(title, series_id, hit_type)
        self.series_id = None
        self.encoded_series_id = None

    def get_seasons(self):
        return self._get_more_data()

    def _get_more_data(self):
        res = Auth.make_get_request(f"https://disney.content.edge.bamgrid.com/svc/content/DmcSeriesBundle/version/5.1/region/{APIConfig.region}/audience/false/maturity/1850/language/{APIConfig.language}/encodedSeriesId/{self.encoded_series_id}")

        res_json = res.json()["data"]["DmcSeriesBundle"]
        self._full_description = res_json["series"]["text"]["description"]["full"]["series"]["default"]["content"]
        self._medium_description = res_json["series"]["text"]["description"]["medium"]["series"]["default"]["content"]
        self._brief_description = res_json["series"]["text"]["description"]["brief"]["series"]["default"]["content"]

        actors, directors, producers, creators = parse_participants(res_json["series"]["participant"])
        self._cast = actors
        self._directors = directors
        self._producers = producers
        self._creators = creators

        seasons = []
        for season_json in res_json["seasons"]["seasons"]:
            season_id = season_json["seasonId"]
            number = season_json["seasonSequenceNumber"]

            season = Season(season_id=season_id, number=number)

            season.release_date = season_json["releases"][0]["releaseDate"]
            season.release_year = season_json["releases"][0]["releaseYear"]
            season.rating = season_json["ratings"][0]["value"]
            season.encoded_series_id = season_json["encodedSeriesId"]
            season.series_id = season_json["seriesId"]

            seasons.append(season)

        return seasons

    def get_related(self):
        res = Auth.make_get_request(f"https://disney.content.edge.bamgrid.com/svc/content/RelatedItems/version/5.1/region/{APIConfig.region}/audience/k-false,l-true/maturity/1850/language/{APIConfig.language}/encodedSeriesId/{self.encoded_series_id}")

        return parse_hits(res.json()["data"]["RelatedItems"]["items"])



from abc import ABC, abstractmethod

from ..models.HitType import HitType


class Hit(ABC):

    def __init__(self, title, hit_id, hit_type: HitType):
        self.title = title
        self.id = hit_id
        self.type = hit_type

        self.release_type = None
        self.release_date = None
        self.release_year = None
        self.images = None
        self.rating = None

        self._full_description = None
        self._medium_description = None
        self._brief_description = None
        self._cast = None
        self._directors = None
        self._producers = None
        self._creators = None
        self.encoded_family_id = None
        self.content_id = None

    @property
    def cast(self):
        if not self._cast:
            self._get_more_data()
        return self._cast

    @property
    def directors(self):
        if not self._directors:
            self._get_more_data()
        return self._directors

    @property
    def producers(self):
        if not self._producers:
            self._get_more_data()
        return self._producers

    @property
    def creators(self):
        if not self._creators:
            self._get_more_data()
        return self._creators

    @property
    def full_description(self):
        if not self._full_description:
            self._get_more_data()
        return self._full_description

    @property
    def medium_description(self):
        if not self._medium_description:
            self._get_more_data()
        return self._medium_description

    @property
    def brief_description(self):
        if not self._brief_description:
            self._get_more_data()
        return self._brief_description

    @abstractmethod
    def _get_more_data(self):
        pass

    def __str__(self):
        return f"[Title={self.title}, Type={self.type.value}]"

    def __repr__(self):
        return self.title



import logging

from ..models.Actor import Actor
from ..models.AudioTrack import AudioTrack
from ..models.HitType import HitType

from ..models.Participant import Participant
from ..models.Profile import Profile
from ..models.Subtitle import Subtitle

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def parse_participants(participants_json):
    def extract_participants(participants_json, participant_type):
        participants = []
        try:
            for participant in participants_json.get(participant_type, []):
                display_name = participant["displayName"]
                order = participant["order"]
                participant_id = participant["participantId"]
                sort_name = participant["sortName"]
                if participant_type == "Actor":
                    character = participant["characterDetails"]["character"]
                    character_id = participant["characterDetails"]["characterId"]
                    participants.append(
                        Actor(character=character, character_id=character_id, display_name=display_name, order=order,
                              actor_id=participant_id,
                              sort_name=sort_name))
                else:
                    participants.append(
                        Participant(display_name=display_name, order=order, participant_id=participant_id,
                                    sort_name=sort_name))
        except KeyError:
            pass
        return participants

    actors = extract_participants(participants_json, "Actor")
    directors = extract_participants(participants_json, "Director")
    producers = extract_participants(participants_json, "Producers")
    creators = extract_participants(participants_json, "Created By")

    return actors, directors, producers, creators


def parse_profile(profile_json):
    profile_id = profile_json["id"]
    name = profile_json["name"]
    kids_mode = profile_json["attributes"]["kidsModeEnabled"]
    is_default = profile_json["attributes"]["isDefault"]
    av_id = profile_json["attributes"]["avatar"]["id"]
    av_user_selected = profile_json["attributes"]["avatar"]["userSelected"]
    prf = Profile(profile_id=profile_id, name=name, kids_mode=kids_mode, is_default=is_default, av_id=av_id,
                  av_user_selected=av_user_selected)

    prf.language_preferences.app = profile_json["attributes"]["languagePreferences"]["appLanguage"]

    prf.language_preferences.playback = profile_json["attributes"]["languagePreferences"]["playbackLanguage"]

    prf.language_preferences.subtitle = profile_json["attributes"]["languagePreferences"]["subtitleLanguage"]
    prf.language_preferences.subs_enabled = profile_json["attributes"]["languagePreferences"]["subtitlesEnabled"]
    return prf


def parse_hits(hits_json, search=False):
    # lazy importing to avoid circular import
    # im dumb and can't design the module structure better
    from ..models.Series import Series
    from ..models.Movie import Movie

    hits = []
    for hit_json in hits_json:
        try:
            if search:
                hit_json = hit_json["hit"]
            contentId = hit_json["contentId"]
            is_movie = True
            try:
                # on default, we think program is a movie
                title = hit_json["text"]["title"]["full"]["program"]["default"]["content"]
            except KeyError:
                # KeyError means it's not a movie so program is a series
                is_movie = False
                title = hit_json["text"]["title"]["full"]["series"]["default"]["content"]
            if is_movie:
                hit_type = HitType.MOVIE
                hit = Movie(title=title, movie_id=contentId, hit_type=hit_type)
                hit.original_language = hit_json["originalLanguage"]

                hit.internal_title = hit_json["internalTitle"]
                hit.format = hit_json["mediaMetadata"]["format"]
                hit.media_id = hit_json["mediaMetadata"]["mediaId"]
                hit.encoded_family_id = hit_json["family"]["encodedFamilyId"]
                hit.length = hit_json["mediaMetadata"]["runtimeMillis"]

            else:
                hit_type = HitType.SERIES
                hit = Series(title=title, series_id=contentId, hit_type=hit_type)

                hit.series_id = hit_json["seriesId"]
                hit.encoded_series_id = hit_json["encodedSeriesId"]

            hit.content_id = hit_json["contentId"]
            hit.images = hit_json["image"]
            hit.release_type = hit_json["releases"][0]["releaseType"]
            hit.release_date = hit_json["releases"][0]["releaseDate"]
            hit.release_year = hit_json["releases"][0]["releaseYear"]

            hit.rating = hit_json["ratings"][0]["value"]

            hits.append(hit)
        except KeyError:
            logger.exception(hit_json)

    return hits


def parse_audio_and_subtitles(string_json, media_id):
    audio_tracks = []
    for audio_track in string_json["mediaMetadata"]["audioTracks"]:
        features = audio_track["features"]
        language = audio_track["language"]
        name = audio_track["renditionName"]

        track_type = audio_track["trackType"]
        audio_track = AudioTrack(features=features, name=name, media_id=media_id, language=language,
                                 track_type=track_type)
        audio_tracks.append(audio_track)

    subtitles = []
    for audio_track in string_json["mediaMetadata"]["captions"]:
        language = audio_track["language"]
        name = audio_track["renditionName"]

        track_type = audio_track["trackType"]
        subtitle = Subtitle(name=name, language=language, media_id=media_id, track_type=track_type)
        subtitles.append(subtitle)

    return audio_tracks, subtitles



import logging
import os

from ..Config import APIConfig
from ..Exceptions import FFmpegException
from ..models.Downloadable import Downloadable
from ..utils.helper import rename_filename

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class AudioTrack(Downloadable):

    def __init__(self, language, name, media_id, track_type, features):
        super().__init__(media_id)
        self.language = language
        self.name = name
        self.track_type = track_type
        self.features = features

    def __str__(self):
        return f"[{self.name}, {self.language}]"

    def __repr__(self):
        return self.name

    def _ffmpeg_download(self, url, path):
        if os.system(
                f'ffmpeg -protocol_whitelist file,http,https,tcp,tls,crypto -y -i "{url}" -c copy "{path}" -preset ultrafast -loglevel warning -hide_banner -stats') != 0:
            raise FFmpegException()

    def download(self, name=None, quality="max"):
        if not name:
            name = self.media_id
        name = rename_filename(name)
        m3u8_url = self._get_m3u8_url(self.media_id)

        _, audio = self._parse_m3u(m3u8_url, self.name, quality)
        path = os.path.join(APIConfig.default_path, name + audio["extension"])
        self._ffmpeg_download(audio["url"], path)

        return path



from enum import Enum


class Language(Enum):
    ALBANIAN = "sq"
    ARABIC = "ar"
    CATALAN = "ca"
    DUTCH = "nl"
    BOSNIAN = "bs"
    BULGARIAN = "bg"
    CROATIAN = "hr"
    CZECH = "cs"
    DANISH = "da"
    ESTONIAN = "et"
    FINNISH = "fi"
    FRENCH = "fr"
    GERMAN = "de"
    GREEK = "el"
    HUNGARIAN = "hu"
    ICELANDIC = "is"
    IRISH = "ga"
    HEBREW = "he"
    ITALIAN = "it"
    SERBIAN = "sr"
    LITHUANIAN = "lt"
    MALTESE = "mt"
    MACEDONIAN = "mk"
    NORWEGIAN = "no"
    POLISH = "pl"
    PORTUGUESE = "pt"
    ROMANIAN = "ro"
    SLOVAK = "sk"
    SLOVENIAN = "sl"
    ENGLISH = "en"
    SPANISH = "es"
    SWEDISH = "sv"
    TURKISH = "tr"



from ..Auth import Auth
from ..Config import APIConfig


class Avatar:
    def __init__(self, avatar_id, user_selected):
        self.id = avatar_id
        self.user_selected = user_selected

        self._name = None
        self._image = None

    @property
    def image(self):
        if not self._image:
            self._get_more_data()
        return self._image

    @property
    def name(self):
        if not self._name:
            self._get_more_data()
        return self._name

    def _get_more_data(self):
        res = Auth.make_get_request(
            f"https://disney.content.edge.bamgrid.com/svc/content/Avatars/version/5.1/region/{APIConfig.region}/audience/k-false,l-true/maturity/1850/language/{APIConfig.language}/avatarIds/{self.id}")
        res_json = res.json()["data"]["Avatars"]["avatars"][0]
        self._image = res_json["image"]["tile"]["1.00"]["avatar"]["default"]["url"]

        self._name = res_json["text"]["title"]["full"]["avatar"]["default"]["content"]

    def __str__(self):
        return f"Avatar({self.id})"

    def __repr__(self):
        return f"Avatar({self.id})"


"https://prod-ripcut-delivery.disney-plus.net/v1/variant/disney/E6BDC816F7F0F83E882316B59622B1041A7DAB53FDE62AA920BCD7C1230AFCD2"



from ..Auth import Auth
from ..Config import APIConfig
from ..models.Hit import Hit
from ..models.HitType import HitType
from ..utils.parser import parse_participants, parse_hits, parse_audio_and_subtitles


class Movie(Hit):

    def __init__(self, title, movie_id, hit_type: HitType):
        super().__init__(title, movie_id, hit_type)
        self.length = None
        self.media_id = None
        self.format = None
        self._subtitles = None
        self._audio_tracks = None
        self.internal_title = None
        self.original_language = None

    @property
    def subtitles(self):
        if not self._subtitles:
            self._get_more_data()
        return self._subtitles

    @property
    def audio_tracks(self):
        if not self._audio_tracks:
            self._get_more_data()
        return self._audio_tracks

    def _get_more_data(self):
        res = Auth.make_get_request(f"https://disney.content.edge.bamgrid.com/svc/content/DmcVideoBundle/version/5.1/region/{APIConfig.region}/audience/k-false,l-true/maturity/1850/language/{APIConfig.language}/encodedFamilyId/{self.encoded_family_id}")
        res_json = res.json()["data"]["DmcVideoBundle"]
        self._brief_description = res_json["video"]["text"]["description"]["brief"]["program"]["default"]["content"]
        self._medium_description = res_json["video"]["text"]["description"]["medium"]["program"]["default"]["content"]
        self._full_description = res_json["video"]["text"]["description"]["full"]["program"]["default"]["content"]
        self.images = res_json["video"]["image"]
        actors, directors, producers, creators = parse_participants(res_json["video"]["participant"])
        self._cast = actors
        self._directors = directors
        self._producers = producers
        self._creators = creators

        audio_tracks, subtitles = parse_audio_and_subtitles(res_json["video"], self.media_id)

        self._subtitles = subtitles
        self._audio_tracks = audio_tracks

    def get_related(self):
        res = Auth.make_get_request(f"https://disney.content.edge.bamgrid.com/svc/content/RelatedItems/version/5.1/region/{APIConfig.region}/audience/k-false,l-true/maturity/1850/language/{APIConfig.language}/encodedFamilyId/{self.encoded_family_id}")
        return parse_hits(res.json()["data"]["RelatedItems"]["items"])



import os
from abc import ABC, abstractmethod
from urllib.parse import urljoin

import m3u8

from ..Auth import Auth


class Downloadable(ABC):

    def __init__(self, media_id):
        self.media_id = media_id

    @abstractmethod
    def download(self):
        pass

    def _get_m3u8_url(self, media_id):

        playback_url = f"https://disney.playback.edge.bamgrid.com/media/{media_id}/scenarios/tvs-drm-cbcs"

        json_data = {
            'playback': {
                'attributes': {
                    'resolution': {
                        'max': [
                            '1920x1080',
                        ],
                    },
                    'protocol': 'HTTPS',
                    'assetInsertionStrategy': 'SGAI',
                    'playbackInitiationContext': 'ONLINE',
                    'frameRates': [
                        60,
                    ],
                    'slugDuration': 'SLUG_500_MS',
                }
            },
        }
        res = Auth.make_post_request(playback_url, json=json_data)

        return res.json()['stream']['sources'][0]['complete']["url"]

    def _parse_m3u(self, m3u_link, name, quality):
        base_url = os.path.dirname(m3u_link)
        subtitle = {}
        audio = {}
        playlists = m3u8.load(uri=m3u_link).playlists

        quality_list = [
            playlist.stream_info.bandwidth for playlist in playlists]

        if quality.lower() == "max":
            quality = quality_list.index(max(quality_list))
        else:
            quality = quality_list.index(min(quality_list))

        for media in playlists[quality].media:
            if media.name == name:

                if media.type == 'SUBTITLES' and media.group_id == 'sub-main':

                    subtitle = {'m3u8_url': urljoin(media.base_uri, media.uri), 'lang': media.language, 'urls': []}

                    segments = m3u8.load(subtitle['m3u8_url'])
                    for uri in segments.files:
                        subtitle['urls'].append(urljoin(segments.base_uri, uri))
                        break

        for media in playlists[quality].media:
            if media.name == name:
                if media.type == 'AUDIO':
                    audio = {'url': f'{base_url}/{media.uri}'}
                    if media.group_id == 'eac-3':
                        audio['extension'] = '.eac3'
                    elif media.group_id in ['aac-128k', 'aac-64k']:
                        audio['url'] = f'{base_url}/{media.uri}'
                        audio['extension'] = '.aac'
                    audio['lang'] = media.language
                    break
        return subtitle, audio



import logging
from typing import Union, Dict


from .Auth import Auth
from .Config import APIConfig
from .models.Account import Account
from .models.Language import Language
from .models.Movie import Movie
from .models.ProgramType import MovieType, SeriesType
from .models.Rating import Rating
from .models.Series import Series
from .utils.parser import parse_hits, parse_profile

logging.basicConfig()

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class DisneyAPI:
    def __init__(self, email: str, password: str, proxies: Dict = False, force_login: bool = False):

        auth = Auth(email=email, password=password, proxies=proxies, force_login=force_login)
        auth.get_auth_token()

        self.device_id = None
        self.device_platform = None
        self.sessions_id = None

        self.account: Account = None
        self._account_init()
        self._session_init()

        APIConfig.region = "en" if self.account is None else self.account.country
        APIConfig.language = APIConfig.region

    def search(self, query: str, rating: Rating = Rating.ADULT):
        res = Auth.make_get_request(
            f"https://disney.content.edge.bamgrid.com/svc/search/disney/version/5.1/region/{APIConfig.region}/audience/k-false,l-true/maturity/{rating.value}/language/{APIConfig.language}/queryType/ge/pageSize/45/query/{query}")
        return parse_hits(res.json()["data"]["search"]["hits"], True)

    def search_movies(self, query: str, rating: Rating = Rating.ADULT):

        return [item for item in self.search(query, rating) if isinstance(item, Movie)]

    def search_series(self, query: str, rating: Rating = Rating.ADULT):

        return [item for item in self.search(query, rating) if isinstance(item, Series)]

    def set_language(self, language: Language):
        APIConfig.language = language.value

    def set_download_path(self, path: str):
        APIConfig.default_path = path

    def get_profiles(self):
        graphql_query = {
            "query": """
                query {
                    me {
                        account {
                            profiles {
                                ...profile
                            }
                        }
                    }
                }
                fragment profile on Profile {
                    id
                    name
                    attributes {
                        avatar {
                            id
                            userSelected
                        }
                        isDefault
                        kidsModeEnabled
                        languagePreferences {
                            appLanguage
                            playbackLanguage
                            subtitleLanguage
                            subtitlesEnabled
                        }
                    }

                }
            """,
            "variables": {}
        }
        res = Auth.make_post_request("https://disney.api.edge.bamgrid.com/v1/public/graphql", json=graphql_query)

        profiles = []
        for profile in res.json()["data"]["me"]["account"]["profiles"]:
            profiles.append(parse_profile(profile))
        return profiles

    def get_active_profile(self):
        """https://global.edge.bamgrid.com/accounts/me/active-profile
        possible duplicate"""

        graphql_query = {
            "query": """
                query {
                    me {
                        account {
                            activeProfile {
                                ...profile
                            }
                        }
                    }
                }
                fragment profile on Profile {
                    id
                    name
                    

                    attributes {
                        avatar {
                            id
                            userSelected
                        }
                        isDefault
                        kidsModeEnabled
                        languagePreferences {
                            playbackLanguage
                            subtitleLanguage
                            subtitlesEnabled
                            appLanguage
                        }
                    }

                }
            """,
            "variables": {}
        }
        res = Auth.make_post_request("https://disney.api.edge.bamgrid.com/v1/public/graphql", json=graphql_query)
        profile = res.json()["data"]["me"]["account"]["activeProfile"]
        return parse_profile(profile)

    def search_program_type(self, program_type: Union[MovieType, SeriesType], rating: Rating = Rating.ADULT):
        res = Auth.make_get_request(f"https://disney.content.edge.bamgrid.com/svc/content/GenericSet/version/6.0/region/{APIConfig.region}/audience/k-false,l-true/maturity/{rating.value}/language/{APIConfig.language}/setId/{program_type.value}/pageSize/100000/page/1")
        return parse_hits(res.json()["data"]["GenericSet"]["items"])

    def set_active_profile(self, profile_id: str, pin: str = None):
        graphql_mutation = {
            "query": """
                mutation switchProfile($input: SwitchProfileInput!) {
                    switchProfile(switchProfile: $input) {
                        account {
                            ...account
                        }
                    }
                }
                fragment account on Account {
                    id
                }
            """,
            "variables": {
                "input": {
                    "profileId": profile_id,
                    "entryPin": pin

                }
            },
            "operationName": "switchProfile"
        }
        res = Auth.make_post_request("https://disney.api.edge.bamgrid.com/v1/public/graphql", json=graphql_mutation)

        APIConfig.token = res.json()["extensions"]["sdk"]["token"]["accessToken"]

    def _account_init(self):
        graphql_query = {
            "query": """
                query {
                    me {
                        account {
                            ...account
                        }
                    }
                }
                fragment account on Account {
                    id
                    attributes {
                        consentPreferences {
                            dataElements {
                                name
                                value
                            }
                        }
                        dssIdentityCreatedAt
                        email
                        emailVerified
                        userVerified
                    }
                }
            """,
            "variables": {}
        }

        res = Auth.make_post_request("https://disney.api.edge.bamgrid.com/v1/public/graphql", json=graphql_query)

        acc_json = res.json()["data"]["me"]["account"]
        account_id = acc_json["id"]
        email = acc_json["attributes"]["email"]
        email_verified = acc_json["attributes"]["emailVerified"]
        created_at = acc_json["attributes"]["dssIdentityCreatedAt"]
        try:
            country = acc_json["attributes"]["consentPreferences"]["dataElements"][0]["value"]
        except IndexError:
            # country seems to be missing for newly create accounts
            country = "en"
            logger.warning("Couldn't set country, fallbacking to default: EN-gb")

        account = Account(account_id=account_id, email=email, created_at=created_at, country=country,
                          is_email_verified=email_verified)
        self.account = account

    def _session_init(self):
        graphql_query = {
            "query": """
                        query {
                            me {
                                activeSession {
                                    ...session
                                }
                            }
                        }
                        fragment session on Session {
                             device {
                                id
                             platform
                             }
                            sessionId
                        }
                    """,
            "variables": {}
        }
        res = Auth.make_post_request("https://disney.api.edge.bamgrid.com/v1/public/graphql", json=graphql_query)

        sess_json = res.json()["data"]["me"]["activeSession"]
        APIConfig.sessionId = sess_json["sessionId"]
        self.device_id = sess_json["device"]["id"]
        self.device_platform = sess_json["device"]["platform"]

    def get_token(self):
        return APIConfig.token



from enum import Enum


class Rating(Enum):
    ADULT = "1850"
    TEEN = "1650"
    KID = "9999"

    AGE18PLUS = "1850"
    AGE16PLUS = "1650"
    AGE14PLUS = "1450"
    AGE12PLUS = "1250"
    AGE9PLUS = "950"
    AGE6PLUS = "650"
    AGE0PLUS = "10"



class Actor:

    def __init__(self, character, character_id, display_name, actor_id, sort_name, order):
        self.character = character
        self.character_id = character_id

        self.display_name = display_name
        self.id = actor_id
        self.sort_name = sort_name
        self.order = order

    def __str__(self):
        return self.display_name

    def __repr__(self):
        return self.display_name



class Episode:
    def __init__(self, content_id: str, number: int, title: str, video_id: str):
        self.content_id = content_id
        self.number = number
        self.title = title
        self.video_id = video_id

        self.season_number = None
        self.internal_title = None
        self.media_id = None
        self.length = None
        self.rating = None
        self.original_language = None
        self.format = None

        self.audio_tracks = None
        self.subtitles = None

        self.brief_description = None
        self.medium_description = None
        self.full_description = None

    def __str__(self):
        return f"[Title={self.title}, Episode={self.number}, Season={self.season_number}]"

    def __repr__(self):
        return self.title



import os
import re
import shutil
from pathlib import Path

import pysubs2

from ..Auth import Auth
from ..Config import APIConfig
from ..models.Downloadable import Downloadable
from ..utils.helper import rename_filename


class Subtitle(Downloadable):

    def __init__(self, language, name, media_id, track_type):
        super().__init__(media_id)
        self.language = language
        self.name = name
        self.track_type = track_type

    def __str__(self):
        return f"Subtitle=[{self.name}, {self.language}]"

    def __repr__(self):
        return self.name

    def is_subtitle(self, file_path, file_format=''):
        extension = Path(file_path).suffix.lower()
        if os.path.isfile(file_path) and Path(file_path).stat().st_size > 0 and extension == ".srt":
            if file_format and file_format != extension:
                return False
            return True
        return False

    def ms_to_timestamp(self, ms: int) -> str:
        max_representable_time = 359999999
        ms = max(ms, 0)
        ms = min(ms, max_representable_time)
        return "%02d:%02d:%02d,%03d" % (pysubs2.time.ms_to_times(ms))

    def convert_list_to_subtitle(self, subs):
        text = ''
        for index, sub in enumerate(subs):
            text = text + str(index + 1) + '\n'
            text = text + self.ms_to_timestamp(sub.start) + ' --> ' + self.ms_to_timestamp(sub.end) + '\n'
            text = text + sub.text.replace('\\n', '\n').replace('\\N', '\n').strip()
            text += '\n\n'

        return pysubs2.ssafile.SSAFile.from_string(text)

    def merge_same_subtitle(self, subs):
        for i, sub in enumerate(subs):
            if i > 0 and sub.text == subs[i - 1].text and sub.start - subs[i - 1].end <= 20:
                subs[i - 1].end = sub.end
                subs.pop(i)
            elif sub.text == '':
                subs.pop(i)
        return subs

    def add_comment(self, subs):
        for sub in subs:
            sub.text = '{\\an8}' + sub.text.strip()
        return subs

    def clean_subs(self, subs):
        for sub in subs:
            text = sub.text
            text = re.sub(r"&rlm;", "", text)
            text = re.sub(r"&lrm;", "", text)
            text = re.sub(r"&amp;", "&", text)
            sub.text = text.strip()
        return subs

    def format_subtitle(self, subs):
        delete_list = []
        for i, sub in enumerate(subs):
            sub.text = re.sub(r'\u200b', '', sub.text)
            sub.text = re.sub(r'\u200e', '', sub.text)
            sub.text = re.sub(r'\u202a', '', sub.text)
            sub.text = re.sub(r'\ufeff', '', sub.text)
            sub.text = re.sub(r'\xa0', ' ', sub.text)

            if sub.text == "":
                delete_list.append(i)

        for i in reversed(delete_list):
            del subs[i]
        return subs

    def _download_subtitle(self, urls, folder_path, name):
        cnt = 0

        for url in urls:
            res = Auth.make_stream_request(url)
            with open(os.path.join(folder_path, f"{str(cnt)}.srt"), 'wb') as file:
                for data in res.iter_content(chunk_size=1024):
                    file.write(data)
            cnt += 1

        self._merge_subtitles(folder_path, name)

    def _merge_subtitles(self, folder_path, name):
        subtitles = []
        for segment in sorted(os.listdir(folder_path)):
            file_path = os.path.join(folder_path, segment)
            if self.is_subtitle(file_path):
                subs = pysubs2.load(file_path)
                subs = self.clean_subs(subs)
                if 'comment' in file_path:
                    self.add_comment(subs)
                subtitles += subs
        subs = self.convert_list_to_subtitle(subtitles)
        subs = self.merge_same_subtitle(subs)
        subs.sort()
        subs = self.format_subtitle(subs)

        subs.save(os.path.join(APIConfig.default_path, name + ".srt"))
        if os.path.exists(folder_path):
            shutil.rmtree(folder_path)

    def download(self, name=None):
        if not name:
            name = self.media_id
        name = rename_filename(name)
        m3u8_url = self._get_m3u8_url(self.media_id)

        subtitle, _ = self._parse_m3u(m3u8_url, self.name, "min")

        folder_path = os.path.join(APIConfig.default_path, "tmp")
        os.makedirs(folder_path, exist_ok=True)

        self._download_subtitle(subtitle['urls'], folder_path, name)



from ..Auth import Auth
from ..Config import APIConfig
from ..Exceptions import ApiException

from ..models.Episode import Episode
from ..utils.parser import parse_audio_and_subtitles


class Season:
    def __init__(self, season_id: str, number: int):
        self.encoded_series_id = None
        self.series_id = None
        self.release_date = None
        self.release_year = None
        self.rating = None
        self.id = season_id
        self.number = number

    def get_episodes(self):
        res = Auth.make_get_request(f"https://disney.content.edge.bamgrid.com/svc/content/DmcEpisodes/version/5.1/region/{APIConfig.region}/audience/false/maturity/1850/language/{APIConfig.language}/seasonId/{self.id}/pageSize/60/page/1")

        if res.status_code != 200:
            raise ApiException(res)
        episodes = []
        for episode_json in res.json()["data"]["DmcEpisodes"]["videos"]:
            content_id = episode_json["contentId"]
            number = episode_json["episodeSeriesSequenceNumber"]
            title = episode_json["text"]["title"]["full"]["program"]["default"]["content"]
            video_id = episode_json["videoId"]
            episode = Episode(content_id=content_id, number=number, title=title, video_id=video_id)

            episode.season_number = self.number
            episode.internal_title = episode_json["internalTitle"]
            episode.media_id = episode_json["mediaMetadata"]["mediaId"]
            episode.original_language = episode_json["originalLanguage"]
            episode.length = episode_json["mediaMetadata"]["runtimeMillis"]
            episode.format = episode_json["mediaMetadata"]["format"]
            episode.rating = episode_json["ratings"][0]["value"]
            episode.brief_description = episode_json["text"]["description"]["brief"]["program"]["default"]["content"]
            episode.medium_description = episode_json["text"]["description"]["medium"]["program"]["default"]["content"]
            episode.full_description = episode_json["text"]["description"]["full"]["program"]["default"]["content"]

            audio_tracks, captions = parse_audio_and_subtitles(episode_json, episode.media_id)

            episode.subtitles = captions
            episode.audio_tracks = audio_tracks
            episodes.append(episode)
        return episodes

    def __str__(self):
        return f"[Id={self.id}, Number={self.number}]"

    def __repr__(self):
        return str(self.number)



from ..models.Avatar import Avatar
from ..models.LanguagePreferences import LanguagePreferences


class Profile:
    def __init__(self, profile_id, name, kids_mode, is_default, av_id, av_user_selected):
        self.id = profile_id
        self.name = name
        self.kids_mode = kids_mode
        self.avatar = Avatar(avatar_id=av_id, user_selected=av_user_selected)
        self.is_default = is_default
        self.language_preferences = LanguagePreferences()

    def __str__(self):
        return f"Profile({self.name})"

    def __repr__(self):
        return self.name





class Participant:
    def __init__(self, display_name, participant_id, sort_name, order):
        self.display_name = display_name
        self.id = participant_id
        self.sort_name = sort_name
        self.order = order

    def __str__(self):
        return self.display_name

    def __repr__(self):
        return self.display_name
    


import json
import logging
import re
from datetime import datetime
from json import JSONDecodeError

from .Config import APIConfig
from .Exceptions import AuthException, ApiException, GraphqlException
from .utils.helper import update_file

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class Auth:
    def __init__(self, email, password, proxies, force_login):
        self._email = email
        self._force_login = force_login
        self._password = password
        self._web_page = 'https://www.disneyplus.com/login'
        self._devices_url = "https://disney.api.edge.bamgrid.com/devices"
        self._login_url = 'https://disney.api.edge.bamgrid.com/idp/login'
        self._token_url = "https://disney.api.edge.bamgrid.com/token"
        self._grant_url = 'https://disney.api.edge.bamgrid.com/accounts/grant'

        if proxies:
            APIConfig.session.proxies.update(proxies)
        APIConfig.auth = self

    def _client_api_key(self):
        res = APIConfig.session.get(self._web_page)
        match = re.search("window.server_path = ({.*});", res.text)
        janson = json.loads(match.group(1))
        clientapikey = janson["sdk"]["clientApiKey"]
        return clientapikey

    def _assertion(self, client_apikey):

        postdata = {
            "applicationRuntime": "firefox",
            "attributes": {},
            "deviceFamily": "browser",
            "deviceProfile": "macosx"
        }

        header = {"authorization": f"Bearer {client_apikey}", "Origin": "https://www.disneyplus.com"}
        res = APIConfig.session.post(url=self._devices_url, headers=header, json=postdata)

        if not res.ok:
            raise AuthException(res)

        assertion = res.json()["assertion"]

        return assertion

    def _access_token(self, client_apikey, assertion):

        header = {"authorization": f"Bearer {client_apikey}", "Origin": "https://www.disneyplus.com"}

        post_date = {
            "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
            "latitude": "0",
            "longitude": "0",
            "platform": "browser",
            "subject_token": assertion,
            "subject_token_type": "urn:bamtech:params:oauth:token-type:device"
        }

        res = APIConfig.session.post(url=self._token_url, headers=header, data=post_date)

        if not res.ok:
            raise AuthException(res)

        access_token = res.json()["access_token"]
        return access_token

    def _login(self, access_token):
        headers = {
            'accept': 'application/json; charset=utf-8',
            'authorization': "Bearer {}".format(access_token),
            'content-type': 'application/json; charset=UTF-8',
            'Origin': 'https://www.disneyplus.com',
            'Referer': 'https://www.disneyplus.com/login/password',
            'Sec-Fetch-Mode': 'cors',
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/76.0.3809.132 Safari/537.36',
            'x-bamsdk-platform': 'windows',
            'x-bamsdk-version': '3.10',
        }

        data = {'email': self._email, 'password': self._password}
        res = APIConfig.session.post(url=self._login_url, data=json.dumps(data), headers=headers)

        if not res.ok:
            raise AuthException(res)

        id_token = res.json()["id_token"]
        return id_token

    def _grant(self, id_token, access_token):

        headers = {
            'accept': 'application/json; charset=utf-8',
            'authorization': f"Bearer {access_token}",
            'content-type': 'application/json; charset=UTF-8',
            'Origin': 'https://www.disneyplus.com',
            'Referer': 'https://www.disneyplus.com/login/password',
            'Sec-Fetch-Mode': 'cors',
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/76.0.3809.132 Safari/537.36',
            'x-bamsdk-platform': 'windows',
            'x-bamsdk-version': '3.10',
        }

        data = {'id_token': id_token}

        res = APIConfig.session.post(url=self._grant_url, data=json.dumps(data), headers=headers)

        if not res.ok:
            raise AuthException(res)

        assertion = res.json()["assertion"]

        return assertion

    def _final_token(self, subject_token, client_apikey):

        header = {"authorization": f"Bearer {client_apikey}", "Origin": "https://www.disneyplus.com"}

        postdata = {
            "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
            "latitude": "0",
            "longitude": "0",
            "platform": "browser",
            "subject_token": subject_token,
            "subject_token_type": "urn:bamtech:params:oauth:token-type:account"
        }

        res = APIConfig.session.post(url=self._token_url, headers=header, data=postdata)

        if not res.ok:
            raise AuthException(res)
        access_token = res.json()["access_token"]
        expires_in = res.json()["expires_in"]
        refresh_token = res.json()["refresh_token"]

        return access_token, expires_in, refresh_token

    def _get_auth_token_tru_api(self):
        logger.info("Loging using disney's api")
        logger.warning("Loging using Disney's api repeatedly will cause account blocks")
        clientapikey_ = self._client_api_key()
        assertion_ = self._assertion(clientapikey_)

        access_token_ = self._access_token(clientapikey_, assertion_)
        id_token_ = self._login(access_token_)

        user_assertion = self._grant(id_token_, access_token_)
        token, expire, refresh = self._final_token(user_assertion, clientapikey_)

        APIConfig.token = token
        APIConfig.refresh = refresh
        update_file()

    @staticmethod
    def refreshToken():
        with open("token.json", "r", encoding="utf8") as file:
            data = json.load(file)

            refresh = data["refresh"]

            logger.info("Refreshing access token using refresh token from token.json file")
            graph_mutation = {
                "query": "mutation refreshToken($input:RefreshTokenInput!){refreshToken(refreshToken:$input){activeSession{sessionId}}}",
                "variables": {
                    "input": {
                        "refreshToken": refresh
                    }
                },
                "operationName": "refreshToken"
            }
            res = APIConfig.session.post(url="https://disney.api.edge.bamgrid.com/graph/v1/device/graphql",
                                         json=graph_mutation,
                                         headers={"authorization": APIConfig.auth._client_api_key()})
            if not res.ok:
                raise AuthException(res)

            if not res.json()["data"]:
                errors = []
                for error in res.json()["errors"]:
                    errors.append(error)
                logger.info("Couldn't refresh access token with refresh token:\n%s", errors)
                #  trying one last time to get access token via standard login procedure using email and password
                APIConfig.auth._get_auth_token_tru_api()
                return

            res_json = res.json()

            APIConfig.token = res_json["extensions"]["sdk"]["token"]["accessToken"]
            APIConfig.refresh = res_json["extensions"]["sdk"]["token"]["refreshToken"]
            update_file()

    @staticmethod
    def make_request(url, is_post, json=None):
        headers = {
            'accept': 'application/vnd.media-service+json; version=6',
            'User-Agent': "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36",
            'x-bamsdk-platform': "macOS",
            'x-bamsdk-version': '23.1',
            'x-dss-edge-accept': 'vnd.dss.edge+json; version=2',
            'x-dss-feature-filtering': 'true',
            'Origin': 'https://www.disneyplus.com',
            'authorization': APIConfig.token
        }
        if is_post:
            if json:
                res = APIConfig.session.post(url=url, json=json, headers=headers, timeout=10)
            else:
                res = APIConfig.session.post(url=url, headers=headers, timeout=10)
        else:
            if json:
                res = APIConfig.session.get(url=url, json=json, headers=headers, timeout=10)
            else:
                res = APIConfig.session.get(url=url, headers=headers, timeout=10)
        if res.status_code == 401:
            raise AuthException(res)
        if not res.ok:
            raise ApiException(res)

        return res

    @staticmethod
    def make_get_request(url, json=None):

        try:
            response = Auth.make_request(url, False, json)
        except AuthException as e:
            Auth.refreshToken()
            response = Auth.make_request(url, False, json)

        return response

    @staticmethod
    def make_stream_request(url):
        headers = {
            'User-Agent': "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36"}

        res = APIConfig.session.get(url=url, headers=headers, stream=True, timeout=10)
        if res.status_code == 401:
            raise AuthException(res)
        if not res.ok:
            raise ApiException(res)

        return res

    @staticmethod
    def make_post_request(url, json=None):
        try:
            response = Auth.make_request(url, True, json)
        except AuthException:
            Auth.refreshToken()
            response = Auth.make_request(url, True, json)

        # This method is used universally, including for downloads where the API response does not include ["data"],
        # hence it can be safely ignored since we are not utilizing the GraphQL endpoint and don't need to handle their errors.
        try:
            if not response.json()["data"]:
                errors = []
                for error in response.json()["errors"]:
                    errors.append(error)
                raise GraphqlException(errors)
        except KeyError:
            pass
        return response

    def get_auth_token(self):
        if self._force_login:
            self._get_auth_token_tru_api()
            return
        try:
            with open("token.json", "r", encoding="utf8") as file:

                data = json.load(file)
                token = data["token"]
                refresh = data["refresh"]
                expiration_time_str = data["expiration_time"]

                # Convert the expiration time string back to a datetime object
                expiration_time = datetime.strptime(expiration_time_str, "%Y-%m-%d %H:%M:%S")

                # Check if the token is still valid
                current_time = datetime.now()
                if current_time < expiration_time:
                    logger.info("Authenticating using token from token.json file")

                    APIConfig.token = token
                    APIConfig.refresh = refresh
                else:
                    logger.info("Access token is expired, refreshing...")

                    self.refreshToken()

        except (FileNotFoundError, KeyError, JSONDecodeError):

            logger.info("Creating token.json file")
            open("token.json", "a+", encoding="utf8").close()

            self._get_auth_token_tru_api()
